/*
 * livepatch_bsc1210499
 *
 * Fix for CVE-2023-0386, bsc#1210499
 *
 *  Upstream commit:
 *  4f11ada10d0a ("ovl: fail on invalid uid/gid mapping at copy up")
 *
 *  SLE12-SP4, SLE12-SP5 and SLE15-SP1 commit:
 *  Not affected
 *
 *  SLE15-SP2 and -SP3 commit:
 *  Not affected
 *
 *  SLE15-SP4 commit:
 *  c351e670a8166f3d4934893c66bb163e026630eb
 *
 *  Copyright (c) 2023 SUSE
 *  Author: Marcos Paulo de Souza <mpdesouza@suse.com>
 *
 *  Based on the original Linux kernel code. Other copyrights apply.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */

/* klp-ccp: from fs/overlayfs/copy_up.c */
#include <linux/module.h>
#include <linux/fs.h>
#include <linux/slab.h>
#include <linux/xattr.h>
#include <linux/uaccess.h>
#include <linux/sched/signal.h>
#include <linux/cred.h>
#include <linux/namei.h>
#include <linux/ratelimit.h>
/* klp-ccp: from fs/overlayfs/overlayfs.h */
#include <linux/kernel.h>
#include <linux/uuid.h>
#include <linux/fs.h>

/* klp-ccp: from fs/overlayfs/ovl_entry.h */
struct ovl_config {
	char *lowerdir;
	char *upperdir;
	char *workdir;
	bool default_permissions;
	bool redirect_dir;
	bool redirect_follow;
	const char *redirect_mode;
	bool index;
	bool uuid;
	bool nfs_export;
	int xino;
	bool metacopy;
	bool userxattr;
	bool ovl_volatile;
};

struct ovl_fs {
	unsigned int numlayer;
	/* Number of unique fs among layers including upper fs */
	unsigned int numfs;
	const struct ovl_layer *layers;
	struct ovl_sb *fs;
	/* workbasedir is the path at workdir= mount option */
	struct dentry *workbasedir;
	/* workdir is the 'work' directory under workbasedir */
	struct dentry *workdir;
	/* index directory listing overlay inodes by origin file handle */
	struct dentry *indexdir;
	long namelen;
	/* pathnames of lower and upper dirs, for show_options */
	struct ovl_config config;
	/* creds of process who forced instantiation of super block */
	const struct cred *creator_cred;
	bool tmpfile;
	bool noxattr;
	/* Did we take the inuse lock? */
	bool upperdir_locked;
	bool workdir_locked;
	bool share_whiteout;
	/* Traps in ovl inode cache */
	struct inode *workbasedir_trap;
	struct inode *workdir_trap;
	struct inode *indexdir_trap;
	/* -1: disabled, 0: same fs, 1..32: number of unused ino bits */
	int xino_mode;
	/* For allocation of non-persistent inode numbers */
	atomic_long_t last_ino;
	/* Whiteout dentry cache */
	struct dentry *whiteout;
	/* r/o snapshot of upperdir sb's only taken on volatile mounts */
	errseq_t errseq;
};

static inline struct ovl_fs *OVL_FS(struct super_block *sb)
{
	return (struct ovl_fs *)sb->s_fs_info;
}

struct ovl_inode {
	union {
		struct ovl_dir_cache *cache;	/* directory */
		struct inode *lowerdata;	/* regular file */
	};
	const char *redirect;
	u64 version;
	unsigned long flags;
	struct inode vfs_inode;
	struct dentry *__upperdentry;
	struct inode *lower;

	/* synchronize copy up and more */
	struct mutex lock;
};

static inline struct ovl_inode *OVL_I(struct inode *inode)
{
	return container_of(inode, struct ovl_inode, vfs_inode);
}

/* klp-ccp: from fs/overlayfs/overlayfs.h */
#undef pr_fmt
#define pr_fmt(fmt) "overlayfs: " fmt

enum ovl_xattr {
	OVL_XATTR_OPAQUE,
	OVL_XATTR_REDIRECT,
	OVL_XATTR_ORIGIN,
	OVL_XATTR_IMPURE,
	OVL_XATTR_NLINK,
	OVL_XATTR_UPPER,
	OVL_XATTR_METACOPY,
};

enum ovl_inode_flag {
	/* Pure upper dir that may contain non pure upper entries */
	OVL_IMPURE,
	/* Non-merge dir that may contain whiteout entries */
	OVL_WHITEOUTS,
	OVL_INDEX,
	OVL_UPPERDATA,
	/* Inode number will remain constant over copy up. */
	OVL_CONST_INO,
};

struct ovl_fb {
	u8 version;	/* 0 */
	u8 magic;	/* 0xfb */
	u8 len;		/* size of this header + size of fid */
	u8 flags;	/* OVL_FH_FLAG_* */
	u8 type;	/* fid_type of fid */
	uuid_t uuid;	/* uuid of filesystem */
	u32 fid[];	/* file identifier should be 32bit aligned in-memory */
} __packed;

struct ovl_fh {
	u8 padding[3];	/* make sure fb.fid is 32bit aligned */
	union {
		struct ovl_fb fb;
		u8 buf[0];
	};
} __packed;

static const char *const (*klpe_ovl_xattr_table)[][2];
static inline const char *klpr_ovl_xattr(struct ovl_fs *ofs, enum ovl_xattr ox)
{
	return (*klpe_ovl_xattr_table)[ox][ofs->config.userxattr];
}

static inline int ovl_do_link(struct dentry *old_dentry, struct inode *dir,
			      struct dentry *new_dentry)
{
	int err = vfs_link(old_dentry, &init_user_ns, dir, new_dentry, NULL);

	pr_debug("link(%pd2, %pd2) = %i\n", old_dentry, new_dentry, err);
	return err;
}

static inline int klpr_ovl_do_setxattr(struct ovl_fs *ofs, struct dentry *dentry,
				  enum ovl_xattr ox, const void *value,
				  size_t size)
{
	const char *name = klpr_ovl_xattr(ofs, ox);
	int err = vfs_setxattr(&init_user_ns, dentry, name, value, size, 0);
	pr_debug("setxattr(%pd2, \"%s\", \"%*pE\", %zu, 0) = %i\n",
		 dentry, name, min((int)size, 48), value, size, err);
	return err;
}

static inline int klpr_ovl_do_removexattr(struct ovl_fs *ofs, struct dentry *dentry,
				     enum ovl_xattr ox)
{
	const char *name = klpr_ovl_xattr(ofs, ox);
	int err = vfs_removexattr(&init_user_ns, dentry, name);
	pr_debug("removexattr(%pd2, \"%s\") = %i\n", dentry, name, err);
	return err;
}

static inline int ovl_do_rename(struct inode *olddir, struct dentry *olddentry,
				struct inode *newdir, struct dentry *newdentry,
				unsigned int flags)
{
	int err;
	struct renamedata rd = {
		.old_mnt_userns	= &init_user_ns,
		.old_dir 	= olddir,
		.old_dentry 	= olddentry,
		.new_mnt_userns	= &init_user_ns,
		.new_dir 	= newdir,
		.new_dentry 	= newdentry,
		.flags 		= flags,
	};

	pr_debug("rename(%pd2, %pd2, 0x%x)\n", olddentry, newdentry, flags);
	err = vfs_rename(&rd);
	if (err) {
		pr_debug("...rename(%pd2, %pd2, ...) = %i\n",
			 olddentry, newdentry, err);
	}
	return err;
}

static inline struct dentry *ovl_do_tmpfile(struct dentry *dentry, umode_t mode)
{
	struct dentry *ret = vfs_tmpfile(&init_user_ns, dentry, mode, 0);
	int err = PTR_ERR_OR_ZERO(ret);

	pr_debug("tmpfile(%pd2, 0%o) = %i\n", dentry, mode, err);
	return ret;
}

static struct dentry *(*klpe_ovl_workdir)(struct dentry *dentry);

static struct dentry *(*klpe_ovl_indexdir)(struct super_block *sb);

static void (*klpe_ovl_path_upper)(struct dentry *dentry, struct path *path);
static void (*klpe_ovl_path_lower)(struct dentry *dentry, struct path *path);
static void (*klpe_ovl_path_lowerdata)(struct dentry *dentry, struct path *path);

static struct dentry *(*klpe_ovl_dentry_upper)(struct dentry *dentry);

static bool (*klpe_ovl_dentry_has_upper_alias)(struct dentry *dentry);
static void (*klpe_ovl_dentry_set_upper_alias)(struct dentry *dentry);

static bool (*klpe_ovl_dentry_needs_data_copy_up_locked)(struct dentry *dentry, int flags);

static void (*klpe_ovl_set_upperdata)(struct inode *inode);

static void (*klpe_ovl_inode_update)(struct inode *inode, struct dentry *upperdentry);

static int (*klpe_ovl_copy_up_start)(struct dentry *dentry, int flags);
static void (*klpe_ovl_copy_up_end)(struct dentry *dentry);

static int (*klpe_ovl_set_impure)(struct dentry *dentry, struct dentry *upperdentry);

static bool (*klpe_ovl_need_index)(struct dentry *dentry);

static inline void ovl_set_flag(unsigned long flag, struct inode *inode)
{
	set_bit(flag, &OVL_I(inode)->flags);
}

static inline bool ovl_test_flag(unsigned long flag, struct inode *inode)
{
	return test_bit(flag, &OVL_I(inode)->flags);
}

static int (*klpe_ovl_get_index_name)(struct ovl_fs *ofs, struct dentry *origin,
		       struct qstr *name);

static int (*klpe_ovl_set_nlink_upper)(struct dentry *dentry);
static int (*klpe_ovl_set_nlink_lower)(struct dentry *dentry);

struct ovl_cattr {
	dev_t rdev;
	umode_t mode;
	const char *link;
	struct dentry *hardlink;
};

#define OVL_CATTR(m) (&(struct ovl_cattr) { .mode = (m) })

static int (*klpe_ovl_cleanup)(struct inode *dir, struct dentry *dentry);

static struct dentry *(*klpe_ovl_create_temp)(struct dentry *workdir, struct ovl_cattr *attr);

static struct ovl_fh *(*klpe_ovl_encode_real_fh)(struct ovl_fs *ofs, struct dentry *real,
				  bool is_upper);

/* klp-ccp: from fs/overlayfs/copy_up.c */
static int (*klpe_ovl_copy_up_data)(struct ovl_fs *ofs, struct path *old,
			    struct path *new, loff_t len);

static int (*klpe_ovl_set_timestamps)(struct dentry *upperdentry, struct kstat *stat);

static int klpr_ovl_set_upper_fh(struct ovl_fs *ofs, struct dentry *upper,
			    struct dentry *index)
{
	const struct ovl_fh *fh;
	int err;

	fh = (*klpe_ovl_encode_real_fh)(ofs, upper, true);
	if (IS_ERR(fh))
		return PTR_ERR(fh);

	err = klpr_ovl_do_setxattr(ofs, index, OVL_XATTR_UPPER, fh->buf, fh->fb.len);

	kfree(fh);
	return err;
}

static int klpr_ovl_create_index(struct dentry *dentry, struct dentry *origin,
			    struct dentry *upper)
{
	struct ovl_fs *ofs = OVL_FS(dentry->d_sb);
	struct dentry *indexdir = (*klpe_ovl_indexdir)(dentry->d_sb);
	struct inode *dir = d_inode(indexdir);
	struct dentry *index = NULL;
	struct dentry *temp = NULL;
	struct qstr name = { };
	int err;

	/*
	 * For now this is only used for creating index entry for directories,
	 * because non-dir are copied up directly to index and then hardlinked
	 * to upper dir.
	 *
	 * TODO: implement create index for non-dir, so we can call it when
	 * encoding file handle for non-dir in case index does not exist.
	 */
	if (WARN_ON(!d_is_dir(dentry)))
		return -EIO;

	/* Directory not expected to be indexed before copy up */
	if (WARN_ON(ovl_test_flag(OVL_INDEX, d_inode(dentry))))
		return -EIO;

	err = (*klpe_ovl_get_index_name)(ofs, origin, &name);
	if (err)
		return err;

	temp = (*klpe_ovl_create_temp)(indexdir, OVL_CATTR(S_IFDIR | 0));
	err = PTR_ERR(temp);
	if (IS_ERR(temp))
		goto free_name;

	err = klpr_ovl_set_upper_fh(ofs, upper, temp);
	if (err)
		goto out;

	index = lookup_one_len(name.name, indexdir, name.len);
	if (IS_ERR(index)) {
		err = PTR_ERR(index);
	} else {
		err = ovl_do_rename(dir, temp, dir, index, 0);
		dput(index);
	}
out:
	if (err)
		(*klpe_ovl_cleanup)(dir, temp);
	dput(temp);
free_name:
	kfree(name.name);
	return err;
}

struct ovl_copy_up_ctx {
	struct dentry *parent;
	struct dentry *dentry;
	struct path lowerpath;
	struct kstat stat;
	struct kstat pstat;
	const char *link;
	struct dentry *destdir;
	struct qstr destname;
	struct dentry *workdir;
	bool origin;
	bool indexed;
	bool metacopy;
};

static int klpr_ovl_link_up(struct ovl_copy_up_ctx *c)
{
	int err;
	struct dentry *upper;
	struct dentry *upperdir = (*klpe_ovl_dentry_upper)(c->parent);
	struct inode *udir = d_inode(upperdir);

	/* Mark parent "impure" because it may now contain non-pure upper */
	err = (*klpe_ovl_set_impure)(c->parent, upperdir);
	if (err)
		return err;

	err = (*klpe_ovl_set_nlink_lower)(c->dentry);
	if (err)
		return err;

	inode_lock_nested(udir, I_MUTEX_PARENT);
	upper = lookup_one_len(c->dentry->d_name.name, upperdir,
			       c->dentry->d_name.len);
	err = PTR_ERR(upper);
	if (!IS_ERR(upper)) {
		err = ovl_do_link((*klpe_ovl_dentry_upper)(c->dentry), udir, upper);
		dput(upper);

		if (!err) {
			/* Restore timestamps on parent (best effort) */
			(*klpe_ovl_set_timestamps)(upperdir, &c->pstat);
			(*klpe_ovl_dentry_set_upper_alias)(c->dentry);
		}
	}
	inode_unlock(udir);
	if (err)
		return err;

	err = (*klpe_ovl_set_nlink_upper)(c->dentry);

	return err;
}

static int (*klpe_ovl_copy_up_inode)(struct ovl_copy_up_ctx *c, struct dentry *temp);

struct ovl_cu_creds {
	const struct cred *old;
	struct cred *new;
};

static int (*klpe_ovl_prep_cu_creds)(struct dentry *dentry, struct ovl_cu_creds *cc);

static void ovl_revert_cu_creds(struct ovl_cu_creds *cc)
{
	if (cc->new) {
		revert_creds(cc->old);
		put_cred(cc->new);
	}
}

static int klpr_ovl_copy_up_workdir(struct ovl_copy_up_ctx *c)
{
	struct inode *inode;
	struct inode *udir = d_inode(c->destdir), *wdir = d_inode(c->workdir);
	struct dentry *temp, *upper;
	struct ovl_cu_creds cc;
	int err;
	struct ovl_cattr cattr = {
		/* Can't properly set mode on creation because of the umask */
		.mode = c->stat.mode & S_IFMT,
		.rdev = c->stat.rdev,
		.link = c->link
	};

	/* workdir and destdir could be the same when copying up to indexdir */
	err = -EIO;
	if (lock_rename(c->workdir, c->destdir) != NULL)
		goto unlock;

	err = (*klpe_ovl_prep_cu_creds)(c->dentry, &cc);
	if (err)
		goto unlock;

	temp = (*klpe_ovl_create_temp)(c->workdir, &cattr);
	ovl_revert_cu_creds(&cc);

	err = PTR_ERR(temp);
	if (IS_ERR(temp))
		goto unlock;

	err = (*klpe_ovl_copy_up_inode)(c, temp);
	if (err)
		goto cleanup;

	if (S_ISDIR(c->stat.mode) && c->indexed) {
		err = klpr_ovl_create_index(c->dentry, c->lowerpath.dentry, temp);
		if (err)
			goto cleanup;
	}

	upper = lookup_one_len(c->destname.name, c->destdir, c->destname.len);
	err = PTR_ERR(upper);
	if (IS_ERR(upper))
		goto cleanup;

	err = ovl_do_rename(wdir, temp, udir, upper, 0);
	dput(upper);
	if (err)
		goto cleanup;

	if (!c->metacopy)
		(*klpe_ovl_set_upperdata)(d_inode(c->dentry));
	inode = d_inode(c->dentry);
	(*klpe_ovl_inode_update)(inode, temp);
	if (S_ISDIR(inode->i_mode))
		ovl_set_flag(OVL_WHITEOUTS, inode);
unlock:
	unlock_rename(c->workdir, c->destdir);

	return err;

cleanup:
	(*klpe_ovl_cleanup)(wdir, temp);
	dput(temp);
	goto unlock;
}

static int klpr_ovl_copy_up_tmpfile(struct ovl_copy_up_ctx *c)
{
	struct inode *udir = d_inode(c->destdir);
	struct dentry *temp, *upper;
	struct ovl_cu_creds cc;
	int err;

	err = (*klpe_ovl_prep_cu_creds)(c->dentry, &cc);
	if (err)
		return err;

	temp = ovl_do_tmpfile(c->workdir, c->stat.mode);
	ovl_revert_cu_creds(&cc);

	if (IS_ERR(temp))
		return PTR_ERR(temp);

	err = (*klpe_ovl_copy_up_inode)(c, temp);
	if (err)
		goto out_dput;

	inode_lock_nested(udir, I_MUTEX_PARENT);

	upper = lookup_one_len(c->destname.name, c->destdir, c->destname.len);
	err = PTR_ERR(upper);
	if (!IS_ERR(upper)) {
		err = ovl_do_link(temp, udir, upper);
		dput(upper);
	}
	inode_unlock(udir);

	if (err)
		goto out_dput;

	if (!c->metacopy)
		(*klpe_ovl_set_upperdata)(d_inode(c->dentry));
	(*klpe_ovl_inode_update)(d_inode(c->dentry), temp);

	return 0;

out_dput:
	dput(temp);
	return err;
}

static int klpr_ovl_do_copy_up(struct ovl_copy_up_ctx *c)
{
	int err;
	struct ovl_fs *ofs = OVL_FS(c->dentry->d_sb);
	bool to_index = false;

	/*
	 * Indexed non-dir is copied up directly to the index entry and then
	 * hardlinked to upper dir. Indexed dir is copied up to indexdir,
	 * then index entry is created and then copied up dir installed.
	 * Copying dir up to indexdir instead of workdir simplifies locking.
	 */
	if ((*klpe_ovl_need_index)(c->dentry)) {
		c->indexed = true;
		if (S_ISDIR(c->stat.mode))
			c->workdir = (*klpe_ovl_indexdir)(c->dentry->d_sb);
		else
			to_index = true;
	}

	if (S_ISDIR(c->stat.mode) || c->stat.nlink == 1 || to_index)
		c->origin = true;

	if (to_index) {
		c->destdir = (*klpe_ovl_indexdir)(c->dentry->d_sb);
		err = (*klpe_ovl_get_index_name)(ofs, c->lowerpath.dentry, &c->destname);
		if (err)
			return err;
	} else if (WARN_ON(!c->parent)) {
		/* Disconnected dentry must be copied up to index dir */
		return -EIO;
	} else {
		/*
		 * Mark parent "impure" because it may now contain non-pure
		 * upper
		 */
		err = (*klpe_ovl_set_impure)(c->parent, c->destdir);
		if (err)
			return err;
	}

	/* Should we copyup with O_TMPFILE or with workdir? */
	if (S_ISREG(c->stat.mode) && ofs->tmpfile)
		err = klpr_ovl_copy_up_tmpfile(c);
	else
		err = klpr_ovl_copy_up_workdir(c);
	if (err)
		goto out;

	if (c->indexed)
		ovl_set_flag(OVL_INDEX, d_inode(c->dentry));

	if (to_index) {
		/* Initialize nlink for copy up of disconnected dentry */
		err = (*klpe_ovl_set_nlink_upper)(c->dentry);
	} else {
		struct inode *udir = d_inode(c->destdir);

		/* Restore timestamps on parent (best effort) */
		inode_lock(udir);
		(*klpe_ovl_set_timestamps)(c->destdir, &c->pstat);
		inode_unlock(udir);

		(*klpe_ovl_dentry_set_upper_alias)(c->dentry);
	}

out:
	if (to_index)
		kfree(c->destname.name);
	return err;
}

static bool ovl_need_meta_copy_up(struct dentry *dentry, umode_t mode,
				  int flags)
{
	struct ovl_fs *ofs = dentry->d_sb->s_fs_info;

	if (!ofs->config.metacopy)
		return false;

	if (!S_ISREG(mode))
		return false;

	if (flags && ((OPEN_FMODE(flags) & FMODE_WRITE) || (flags & O_TRUNC)))
		return false;

	return true;
}

static ssize_t ovl_getxattr(struct dentry *dentry, char *name, char **value)
{
	ssize_t res;
	char *buf;

	res = vfs_getxattr(&init_user_ns, dentry, name, NULL, 0);
	if (res == -ENODATA || res == -EOPNOTSUPP)
		res = 0;

	if (res > 0) {
		buf = kzalloc(res, GFP_KERNEL);
		if (!buf)
			return -ENOMEM;

		res = vfs_getxattr(&init_user_ns, dentry, name, buf, res);
		if (res < 0)
			kfree(buf);
		else
			*value = buf;
	}
	return res;
}

static int klpr_ovl_copy_up_meta_inode_data(struct ovl_copy_up_ctx *c)
{
	struct ovl_fs *ofs = OVL_FS(c->dentry->d_sb);
	struct path upperpath, datapath;
	int err;
	char *capability = NULL;
	ssize_t cap_size;

	(*klpe_ovl_path_upper)(c->dentry, &upperpath);
	if (WARN_ON(upperpath.dentry == NULL))
		return -EIO;

	(*klpe_ovl_path_lowerdata)(c->dentry, &datapath);
	if (WARN_ON(datapath.dentry == NULL))
		return -EIO;

	if (c->stat.size) {
		err = cap_size = ovl_getxattr(upperpath.dentry, XATTR_NAME_CAPS,
					      &capability);
		if (cap_size < 0)
			goto out;
	}

	err = (*klpe_ovl_copy_up_data)(ofs, &datapath, &upperpath, c->stat.size);
	if (err)
		goto out_free;

	/*
	 * Writing to upper file will clear security.capability xattr. We
	 * don't want that to happen for normal copy-up operation.
	 */
	if (capability) {
		err = vfs_setxattr(&init_user_ns, upperpath.dentry,
				   XATTR_NAME_CAPS, capability, cap_size, 0);
		if (err)
			goto out_free;
	}


	err = klpr_ovl_do_removexattr(ofs, upperpath.dentry, OVL_XATTR_METACOPY);
	if (err)
		goto out_free;

	(*klpe_ovl_set_upperdata)(d_inode(c->dentry));
out_free:
	kfree(capability);
out:
	return err;
}

int klpp_ovl_copy_up_one(struct dentry *parent, struct dentry *dentry,
			   int flags)
{
	int err;
	DEFINE_DELAYED_CALL(done);
	struct path parentpath;
	struct ovl_copy_up_ctx ctx = {
		.parent = parent,
		.dentry = dentry,
		.workdir = (*klpe_ovl_workdir)(dentry),
	};

	if (WARN_ON(!ctx.workdir))
		return -EROFS;

	(*klpe_ovl_path_lower)(dentry, &ctx.lowerpath);
	err = vfs_getattr(&ctx.lowerpath, &ctx.stat,
			  STATX_BASIC_STATS, AT_STATX_SYNC_AS_STAT);
	if (err)
		return err;

	if (!kuid_has_mapping(current_user_ns(), ctx.stat.uid) ||
	    !kgid_has_mapping(current_user_ns(), ctx.stat.gid))
		return -EOVERFLOW;

	ctx.metacopy = ovl_need_meta_copy_up(dentry, ctx.stat.mode, flags);

	if (parent) {
		(*klpe_ovl_path_upper)(parent, &parentpath);
		ctx.destdir = parentpath.dentry;
		ctx.destname = dentry->d_name;

		err = vfs_getattr(&parentpath, &ctx.pstat,
				  STATX_ATIME | STATX_MTIME,
				  AT_STATX_SYNC_AS_STAT);
		if (err)
			return err;
	}

	/* maybe truncate regular file. this has no effect on dirs */
	if (flags & O_TRUNC)
		ctx.stat.size = 0;

	if (S_ISLNK(ctx.stat.mode)) {
		ctx.link = vfs_get_link(ctx.lowerpath.dentry, &done);
		if (IS_ERR(ctx.link))
			return PTR_ERR(ctx.link);
	}

	err = (*klpe_ovl_copy_up_start)(dentry, flags);
	/* err < 0: interrupted, err > 0: raced with another copy-up */
	if (unlikely(err)) {
		if (err > 0)
			err = 0;
	} else {
		if (!(*klpe_ovl_dentry_upper)(dentry))
			err = klpr_ovl_do_copy_up(&ctx);
		if (!err && parent && !(*klpe_ovl_dentry_has_upper_alias)(dentry))
			err = klpr_ovl_link_up(&ctx);
		if (!err && (*klpe_ovl_dentry_needs_data_copy_up_locked)(dentry, flags))
			err = klpr_ovl_copy_up_meta_inode_data(&ctx);
		(*klpe_ovl_copy_up_end)(dentry);
	}
	do_delayed_call(&done);

	return err;
}



#define LP_MODULE "overlay"

#include <linux/kernel.h>
#include <linux/module.h>
#include "livepatch_bsc1210499.h"
#include "../kallsyms_relocs.h"

static struct klp_kallsyms_reloc klp_funcs[] = {
	{ "ovl_cleanup", (void *)&klpe_ovl_cleanup, "overlay" },
	{ "ovl_copy_up_data", (void *)&klpe_ovl_copy_up_data, "overlay" },
	{ "ovl_copy_up_end", (void *)&klpe_ovl_copy_up_end, "overlay" },
	{ "ovl_copy_up_inode", (void *)&klpe_ovl_copy_up_inode, "overlay" },
	{ "ovl_copy_up_start", (void *)&klpe_ovl_copy_up_start, "overlay" },
	{ "ovl_create_temp", (void *)&klpe_ovl_create_temp, "overlay" },
	{ "ovl_dentry_has_upper_alias",
	  (void *)&klpe_ovl_dentry_has_upper_alias, "overlay" },
	{ "ovl_dentry_needs_data_copy_up_locked",
	  (void *)&klpe_ovl_dentry_needs_data_copy_up_locked, "overlay" },
	{ "ovl_dentry_set_upper_alias",
	  (void *)&klpe_ovl_dentry_set_upper_alias, "overlay" },
	{ "ovl_dentry_upper", (void *)&klpe_ovl_dentry_upper, "overlay" },
	{ "ovl_encode_real_fh", (void *)&klpe_ovl_encode_real_fh, "overlay" },
	{ "ovl_get_index_name", (void *)&klpe_ovl_get_index_name, "overlay" },
	{ "ovl_indexdir", (void *)&klpe_ovl_indexdir, "overlay" },
	{ "ovl_inode_update", (void *)&klpe_ovl_inode_update, "overlay" },
	{ "ovl_need_index", (void *)&klpe_ovl_need_index, "overlay" },
	{ "ovl_path_lower", (void *)&klpe_ovl_path_lower, "overlay" },
	{ "ovl_path_lowerdata", (void *)&klpe_ovl_path_lowerdata, "overlay" },
	{ "ovl_path_upper", (void *)&klpe_ovl_path_upper, "overlay" },
	{ "ovl_prep_cu_creds", (void *)&klpe_ovl_prep_cu_creds, "overlay" },
	{ "ovl_set_impure", (void *)&klpe_ovl_set_impure, "overlay" },
	{ "ovl_set_nlink_lower", (void *)&klpe_ovl_set_nlink_lower,
	  "overlay" },
	{ "ovl_set_nlink_upper", (void *)&klpe_ovl_set_nlink_upper,
	  "overlay" },
	{ "ovl_set_timestamps", (void *)&klpe_ovl_set_timestamps, "overlay" },
	{ "ovl_set_upperdata", (void *)&klpe_ovl_set_upperdata, "overlay" },
	{ "ovl_workdir", (void *)&klpe_ovl_workdir, "overlay" },
	{ "ovl_xattr_table", (void *)&klpe_ovl_xattr_table, "overlay" },
};

static int module_notify(struct notifier_block *nb,
			unsigned long action, void *data)
{
	struct module *mod = data;
	int ret;

	if (action != MODULE_STATE_COMING || strcmp(mod->name, LP_MODULE))
		return 0;
	ret = klp_resolve_kallsyms_relocs(klp_funcs, ARRAY_SIZE(klp_funcs));

	WARN(ret, "%s: delayed kallsyms lookup failed. System is broken and can crash.\n",
		__func__);

	return ret;
}

static struct notifier_block module_nb = {
	.notifier_call = module_notify,
	.priority = INT_MIN+1,
};

int livepatch_bsc1210499_init(void)
{
	int ret;
	struct module *mod;

	ret = klp_kallsyms_relocs_init();
	if (ret)
		return ret;

	ret = register_module_notifier(&module_nb);
	if (ret)
		return ret;

	rcu_read_lock_sched();
	mod = (*klpe_find_module)(LP_MODULE);
	if (!try_module_get(mod))
		mod = NULL;
	rcu_read_unlock_sched();

	if (mod) {
		ret = klp_resolve_kallsyms_relocs(klp_funcs,
						ARRAY_SIZE(klp_funcs));
	}

	if (ret)
		unregister_module_notifier(&module_nb);
	module_put(mod);

	return ret;
}

void livepatch_bsc1210499_cleanup(void)
{
	unregister_module_notifier(&module_nb);
}
